'use strict';

const bodyParser = require('body-parser');
const cookieParser = require('cookie-parser');
const express = require('express');
const fs = require('fs');
const os = require('os');
const path = require('path');
const socketio = require('socket.io');
const winston = require('winston');

const ApiBuilder = require('./services/api.builder');
const SwaggerApiParser = require('./services/swagger.parser');

// Express
const application = express();

application.use(bodyParser.json());

application.use(bodyParser.urlencoded({ extended: true }));

application.use(cookieParser());

// API Builder
const apiBuilder = new ApiBuilder();

const folderCreator = (folderPath) => {
    winston.debug(`Validating folder path: ${folderPath}`);
    const folders = folderPath.split(path.sep);
    folders.reduce(
        (agg, current) => {
            agg = agg ? path.join(agg, current) : current;
            if (!fs.existsSync(agg)) {
                fs.mkdirSync(agg);
            }
            return agg;
        },
        os.platform() === 'win32' ? '' : '/'
    );
};

/**
 * <p><code>Function</code> that initializes the spec file parsing, configured and finally initializes
 * the <i>ExpressJS</i> instance that mocks the input spec.</p>
 *
 * @param filePath {string} containing the spec file path on the file system.
 * @param options {JSON} which contains process configuration: output file path, mock file path and port number among others.
 * configuration file.
 */
const thin2mocker = (filePath, options) => {
    const outputFilePath = options.output;
    const mockFilePath = options.mock;
    const portNumber = options.port;
    const logsPath = options.logs;
    const logsLevel = options.level;
    const novaMode = options.nova;

    // Winston initialization
    winston.level = logsLevel || 'debug';

    // Configure file output if neccesary
    if (logsPath) {
        folderCreator(logsPath);
        winston.add(winston.transports.File, { filename: path.join(logsPath, 'mock.log') });
    }

    winston.debug(`Reading Swagger spec file at ${filePath}`);
    winston.debug(`Reading mock spec file at ${mockFilePath ? mockFilePath : 'not specified'}`);

    const applicationPort = portNumber || 4000;

    // Start the process
    fs.readFile(filePath, (err, swaggerSpec) => {
        if (err) {
            winston.error(`Unable to read Swagger file: ${err}`);
        } else {
            SwaggerApiParser.extractApiData(swaggerSpec, mockFilePath)
                .then((apiModel) => {
                    if (outputFilePath) {
                        const targetPath = path.join(outputFilePath, 'output.json');
                        winston.debug(`Generating output file at ${targetPath}`);
                        fs.writeFileSync(targetPath, JSON.stringify(apiModel));
                    }

                    apiBuilder.build(apiModel).then(
                        (configuration) => {
                            const routes = configuration.routes;
                            const sockets = configuration.sockets;
                            winston.debug(
                                `API successfully configured: ${routes.stack.length} routes and ${
                                    sockets.length
                                } sockets`
                            );
                            winston.debug('Configuring CORS');
                            application.use((request, response, next) => {
                                let oneof = false;
                                if (request.headers.origin) {
                                    response.header('Access-Control-Allow-Origin', request.headers.origin);
                                    oneof = true;
                                }
                                if (request.headers['access-control-request-method']) {
                                    response.header(
                                        'Access-Control-Allow-Methods',
                                        request.headers['access-control-request-method']
                                    );
                                    oneof = true;
                                }
                                if (request.headers['access-control-request-headers']) {
                                    response.header(
                                        'Access-Control-Allow-Headers',
                                        request.headers['access-control-request-headers']
                                    );
                                    oneof = true;
                                }
                                if (oneof) {
                                    response.header('Access-Control-Max-Age', 60 * 60 * 24 * 365);
                                }
                                // Allow all preflight OPTIONS requests in order to avoid issues since they don't have
                                // the Authorization header
                                if (request.method === 'OPTIONS') {
                                    response.send(200);
                                } else {
                                    next();
                                }
                            });
                            winston.debug('Initializing express routes.');

                            application.use((request, response, next) => {
                                winston.debug(` ## Received call at ${request.url} `);
                                winston.debug(`  # Headers ${JSON.stringify(request.headers)} `);
                                winston.debug(`  # Cookies ${JSON.stringify(request.cookies)} `);
                                winston.debug(`  # Signed Cookies ${JSON.stringify(request.signedCookies)} `);
                                winston.debug(`  # Params ${JSON.stringify(request.params)} `);
                                winston.debug(`  # Body ${JSON.stringify(request.body)} `);

                                next();
                                winston.debug(` ## Finished call at ${request.url} `);
                            });

                            winston.debug(`API base path: ${apiModel.basePath}`);
                            if (novaMode) {
                                winston.debug('NOVA mode enabled, normalizing base path');
                                apiModel.basePath = apiModel.basePath
                                    .replace(`/${apiModel.uuaa}`, '')
                                    .replace(`/${apiModel.name}`, '')
                                    .replace(`/${apiModel.version}`, '');
                                winston.debug(`NOVA normalized API base path: ${apiModel.basePath}`);
                            }
                            application.use(apiModel.basePath || '', routes);
                            const server = application.listen(applicationPort, () => {
                                winston.debug(`Thin2-Mocker backend listening on port ${applicationPort}!`);
                            });

                            winston.debug('Initializing socket.io sockets.');
                            sockets.forEach((socketData) => {
                                const transports = [];
                                transports.push(socketData.transport);
                                if (socketData.transport === 'websocket') {
                                    transports.push('polling');
                                }

                                const socketConnection = socketio(server, {
                                    transports: transports,
                                    pingInterval: socketData.ping,
                                    path: socketData.url
                                });
                                socketConnection.on('connection', (socket) => {
                                    winston.debug(`Connection on socket ${socketData.url} opened`);
                                    socketData.fn(socket);
                                    socket.on('disconnect', () => {
                                        winston.debug(`Connection on socket ${socketData.url} closed`);
                                    });
                                });
                            });

                            application.on('error', (error) => {
                                winston.error(` ## ERROR # ${error} ## `);
                                throw error;
                            });

                            winston.debug('API successfully initialized.');
                        },
                        (err) => winston.error(`Error while configuring API: ${err}`)
                    );
                })
                .catch((err) => winston.error(err));
        }
    });
};

module.exports = thin2mocker;
