'use strict';

const express = require('express');
const winston = require('winston');

const ParameterValidator = require('../utils/parameter.validator');
const ResponseGenerator = require('../utils/response.generator');

/**
 * <p><code>Builder</code> object that parses the input <i>API</i> model and generates the endpoints,
 * parameter validators and response generators.</p>
 */
class ApiBuilder {
    /**
     * <p><code>Builder</code>'s constructor.</p>
     */
    constructor() {
        this._router = express.Router();
        this._sockets = [];

        this.build.bind(this);
        this._findDefinition.bind(this);
        this._resolveDefinition.bind(this);
        this._resolveParameter.bind(this);
        this._resolveResponse.bind(this);
    }

    /**
     * <p>Method that configures a working <i>API</i> from the given <i>API</i> model.
     *
     * @param apiModel {ApiModel} containing the <i>API</i> configuration data.
     * @returns {Promise} resolves a configured router with the <i>API</i> configurations,
     * rejects in case of error with the error message.
     */
    build(apiModel) {
        return new Promise((resolve, reject) => {
            try {
                winston.debug(`Building API ${apiModel.name}:${apiModel.version}`);
                apiModel.endpoints.forEach((endpoint) => {
                    winston.debug(` - Building endpoint ${endpoint.method.toUpperCase()}: ${endpoint.path}`);

                    // 1. Build parameter validators
                    const validators = [];
                    endpoint.parameters.forEach((parameter) => {
                        const paramName = parameter.name ? parameter.name : parameter.format;
                        winston.debug(`  # Building validator for parameter ${paramName}`);

                        let parameterName;
                        let parameterLocation;
                        let format;

                        const match = /#\/(parameters|definitions)\/([-\w]+)/g.exec(parameter.format);
                        if (match) {
                            const isArray = parameter.format.startsWith('array');
                            const parameterFormat = match[1];
                            const parameterRefName = match[2];
                            // FIXME: 12/06/2018
                            // This OR clause makes me believe we're allocating Swagger's spec format straight
                            // up into the encapsulated model.
                            const param = apiModel[parameterFormat].find(
                                (parameter) =>
                                    parameter.definitionName === parameterRefName || parameter.name === parameterRefName
                            );
                            const resolvedParameter = this._resolveParameter(
                                param,
                                parameterFormat,
                                apiModel.definitions
                            );
                            parameterName = parameter.name || param.name;
                            parameterLocation = parameter.location || param.location;

                            if (isArray) {
                                format = [resolvedParameter];
                            } else {
                                format = resolvedParameter;
                            }
                        } else {
                            const match = /array\[([-\w]+)\]/g.exec(parameter.format);

                            parameterName = parameter.name;
                            parameterLocation = parameter.location;

                            if (match) {
                                format = [match[1]];
                            } else {
                                format = parameter.format;
                            }
                        }

                        const parameterValidator = new ParameterValidator(parameterName, parameterLocation, format);

                        validators.push(parameterValidator);
                    });

                    // 2. Build response generators
                    const responses = [];
                    endpoint.responses.forEach((response) => {
                        winston.debug(`  # Building response generator for response ${response.code}`);

                        // FIXME: How do we treat unformatted responses?
                        let format;
                        if (response.format) {
                            const match = /#\/(responses|definitions)\/([-\w]+)/g.exec(response.format);
                            if (match) {
                                const isArray = response.format.startsWith('array');
                                // These definitions can be found on either responses or definitions
                                const responseFormat = match[1];
                                const responseName = match[2];
                                const resp = apiModel[responseFormat].find(
                                    (response) => response.name === responseName
                                );
                                const resolvedResponse = this._resolveResponse(resp, apiModel.definitions);

                                if (isArray) {
                                    format = [resolvedResponse];
                                } else {
                                    format = resolvedResponse;
                                }
                            } else {
                                let parsedType;
                                const match = /array\[([-\w]+)\]/g.exec(response.format);
                                if (match) {
                                    parsedType = [match[1]];
                                } else {
                                    parsedType = response.format;
                                }
                                format = parsedType;
                            }
                        }

                        const responseGenerator = new ResponseGenerator(response.code, format);

                        responses.push(responseGenerator);
                    });

                    // 3. Generate a ExpressJS endpoint with the generated configuration
                    const defaultEndpoint = (request, response) => {
                        // const requestValidators = validators;
                        validators.forEach((validator) => {
                            const validationResult = validator.validate(request);
                            if (!validationResult) {
                                winston.error(`Parameter ${validator.name} is not valid`);
                            }
                        });

                        // FIXME: For the time being we'll be defaulting to success responses
                        // FIXME: Support an scenario of no response definitions?

                        // const responseGenerators = responses;
                        const successResponse = responses.find((response) => response.code === '200');
                        // If no success response is found, attempt to get default response
                        const defaultResponse = responses.find((response) => response.code === 'default');
                        // If neither are found, just get the first configured response
                        const firstResponse = responses[0];

                        if (successResponse) {
                            successResponse.generate(response);
                        } else if (defaultResponse) {
                            defaultResponse.generate(response);
                        } else {
                            firstResponse.generate(response);
                        }
                    };

                    const endpointFn = endpoint.functionToExec
                        ? endpoint.functionToExec(defaultEndpoint)
                        : defaultEndpoint;

                    switch (endpoint.method) {
                        case 'get':
                            this._router.get(endpoint.path, endpointFn);
                            break;
                        case 'post':
                            this._router.post(endpoint.path, endpointFn);
                            break;
                        case 'put':
                            this._router.put(endpoint.path, endpointFn);
                            break;
                        case 'delete':
                            this._router.delete(endpoint.path, endpointFn);
                            break;
                        default:
                            throw new Error('Unsupported HTTP method.');
                    }
                });

                apiModel.sockets.forEach((socketData) => {
                    winston.debug(` - Building socket: ${socketData.url}`);
                    winston.debug(`  # Transport: ${socketData.transport}`);
                    // Ping will be undefined for non-polling transports
                    if (socketData.ping) {
                        winston.debug(`  # Ping: ${socketData.ping}`);
                    }

                    const payload = this._findDefinition(socketData.format, apiModel.definitions);

                    const socketFn = (socket) => {
                        winston.debug(
                            `Configuring ${socketData.interval} data generation for event ${socketData.event}.`
                        );
                        const intervalId = setInterval(() => {
                            socket.emit(socketData.event, payload);
                        }, socketData.interval);

                        socket.on('disconnect', () => {
                            clearInterval(intervalId);
                        });
                    };

                    const finalSocketFn = socketData.functionToExec ? socketData.functionToExec(socketFn) : socketFn;

                    const socketFormat = {
                        fn: finalSocketFn,
                        url: socketData.url,
                        transport: socketData.transport,
                        ping: socketData.ping
                    };

                    this._sockets.push(socketFormat);
                });

                const resolvedConfiguration = {
                    routes: this._router,
                    sockets: this._sockets
                };
                resolve(resolvedConfiguration);
            } catch (err) {
                reject(err);
            }
        });
    }

    /**
     * <p>Method that finds a <code>Definition</code> by it's reference.</p>
     *
     * @param type {string} containing the type we want to parse.
     * @param definitions {Definitions[]} containing all the available <code>Definitions</code>.
     * @param processedDefinitions {string[]} containing the <code>Definitions</code> that have already been
     * processed in order to avoid infinite recursive loops.
     * @returns {Object} containing the format of the found <code>Definition</code>.
     * @private
     */
    _findDefinition(type, definitions, processedDefinitions) {
        let parsedType;

        const readDefinitions = processedDefinitions || [];

        const match = /#\/definitions\/([-\w]+)/g.exec(type);
        if (match) {
            const isArray = type.startsWith('array');
            const definitionName = match[1];
            const definition = definitions.find((definition) => definition.name === definitionName);
            readDefinitions.push(definitionName);
            const resolvedDefinition = this._resolveDefinition(definition, definitions, readDefinitions);

            if (isArray) {
                parsedType = [resolvedDefinition];
            } else {
                parsedType = resolvedDefinition;
            }
        } else {
            const match = /array\[([-\w]+)\]/g.exec(type);
            if (match) {
                parsedType = [match[1]];
            } else {
                parsedType = type;
            }
        }

        return parsedType;
    }

    /**
     * <p>Method that generates an object using the given <code>Definition</code>'s format.</p>
     *
     * @param definition {Definition} containing the object which format we want to use.
     * @param definitions {Definitions[]} containing all the available <code>Definitions</code>.
     * @param readDefinitions {string[]} containing the <code>Definitions</code> that have already been
     * processed in order to avoid infinite recursive loops.
     * @returns {Object} with an object generated with the given <code>Definition</code> format.
     * @private
     */
    _resolveDefinition(definition, definitions, readDefinitions) {
        // NOTE: Cyclic elements will be resolved at one level only
        let finalFormat;
        const format = definition.format.format;
        if (format) {
            finalFormat = {};
            format.forEach((element) => {
                if (element.extends) {
                    const extendedProps = this._findDefinition(element.extends, definitions, readDefinitions);
                    const props = Object.keys(extendedProps);
                    props.forEach((prop) => {
                        finalFormat[prop] = extendedProps[prop];
                    });
                } else {
                    // Limit the amount of times the same definition can be read to two
                    // A for-loop would be more efficient than this, but we're not really worried about performance here
                    const definitionCount = readDefinitions.filter((defName) => defName === definition.name).length;
                    if (definitionCount >= 2) {
                        finalFormat[element.name] = element.type;
                    } else {
                        finalFormat[element.name] = this._findDefinition(element.type, definitions, readDefinitions);
                    }
                }
            });
        } else {
            finalFormat = definition.format.type;
        }

        return finalFormat;
    }

    /**
     * <p>Method that generates an object using the given <code>Parameter</code>'s format.</p>
     *
     * @param parameter {Parameter} containing the object which format we want to use.
     * @param parameterFormatLocation {string} containing the parameter reference location (definitions or parameters).
     * @param definitions {Definitions[]} containing all the available <code>Definitions</code>.
     * @returns {Object} with an object generated with the given <code>Parameter</code> format.
     * @private
     */
    _resolveParameter(parameter, parameterFormatLocation, definitions) {
        let finalFormat;

        if (parameterFormatLocation === 'definitions') {
            finalFormat = {};
            const format = parameter.format.format;
            format.forEach((element) => {
                const format = this._findDefinition(element.type, definitions);
                finalFormat[element.name] = format;
            });
        } else {
            finalFormat = parameter.format;
        }

        return finalFormat;
    }

    /**
     * <p>Method that generates an object using the given <code>Response</code>'s format.</p>
     *
     * @param response {Response} containing the object which format we want to use.
     * @param definitions {Definitions[]} containing all the available <code>Definitions</code>.
     * @returns {Object} with an object generated with the given <code>Response</code> format.
     * @private
     */
    _resolveResponse(response, definitions) {
        const finalFormat = {};

        const format = response.format.format;
        format.forEach((element) => {
            const format = this._findDefinition(element.type, definitions);
            finalFormat[element.name] = format;
        });

        return finalFormat;
    }
}

module.exports = ApiBuilder;
