'use strict';

const SwaggerParser = require('swagger-parser');
const winston = require('winston');
const fs = require('fs');
const yamljs = require('yaml');

const CONSTANTS = require('./../constants/constants');

const ApiModel = require('./../models/api.model');
const DefinitionModel = require('./../models/definition.model');
const EndpointModel = require('./../models/endpoint.model');
const ParameterModel = require('./../models/parameter.model');
const ResponseModel = require('./../models/response.model');
const SchemaModel = require('./../models/schema.model');
const SocketModel = require('./../models/socket.model');

/**
 * <p><code>Parser</code> object that parses the input <i>Swagger</i> spec file and generates the <i>API</i> model,
 * which will be used to configure the mock <i>API</i> or the code generation.</p>
 */
class SwaggerApiParser {
    /**
     * <p><code>Parser</code>'s constructor.</p>
     */
    constructor() {
        this.mockFileData;
        this.allocateMockData.bind(this);
        this.extractApiData.bind(this);
        this.extractMethodsData.bind(this);
        this.extractMockFileData.bind(this);
        this.extractSocketData.bind(this);
        this.getType.bind(this);
        this.parseArrayItems.bind(this);
        this.parseSchema.bind(this);
        this.parseSwaggerSpec.bind(this);
        this.validateSwaggerSpec.bind(this);
    }

    /**
     * <p>Method that allocates the mock data for the given endpoint and method from
     * either the mock file or the <i>Swagger</i> file function.</p>
     * @param uri {string} endpoint's uri.
     * @param method {string} endpoint's method (get, post, put...).
     * @param pathConfig {Object} contains the uri / method Swagger configuration metadata.
     * @return this method returns a factory function which receives a default handler which
     * will be invoked in case the filters (path/query params) are not matched.
     */
    allocateMockData(uri, method, pathConfig) {
        let factoryFn;

        // 1. Attempt to allocate data mock from file
        if (this.mockFileData && this.mockFileData.endpoints.length > 0) {
            winston.debug(`Searching file mocks for endpoint ${method}:${uri}`);
            const mockData = this.mockFileData.endpoints.filter((data) => data.uri === uri && data.method === method);

            if (mockData.length === 1 && !mockData[0].path) {
                winston.debug(`Found mock data for endpoint ${method}:${uri}`);
                factoryFn = () => (request, response) => {
                    const mockResponse = mockData[0].data;
                    response.json(mockResponse);
                };
            } else if (mockData.length > 0) {
                winston.debug(`Found mock data for endpoint ${method}:${uri}`);
                factoryFn = (defaultHandler) => (request, response) => {
                    const mockResponse = mockData.find((mock) =>
                        mock.path.every((pathArg) => request.params[pathArg.param] === String(pathArg.value))
                    );
                    if (mockResponse) {
                        response.json(mockResponse.data);
                    } else {
                        defaultHandler(request, response);
                    }
                };
            }
        }

        // 2. Attempt to allocate data mock from Swagger spec file
        const fnDataFromFile = pathConfig[method][CONSTANTS.CUSTOM_FN];
        if (fnDataFromFile) {
            const fnSpec = fs.readFileSync(fnDataFromFile, 'utf8');
            factoryFn = () => eval(fnSpec);
        }

        return factoryFn;
    }

    /**
     * <p>Method that extracts the <i>API</i> configuration object from the input <i>Swagger</i> spec file.</p>
     *
     * @param swaggerSpec {string} with the <i>Swagger</i> spec file to parse.
     * @param mockFilePath {string} with the mock data file path.
     * @returns {Promise} resolves the parsed <i>Swagger</i> spec configuration or rejects in case of error.
     */
    extractApiData(swaggerSpec, mockFilePath) {
        return new Promise((resolve, reject) => {
            this.extractMockFileData(mockFilePath)
                .then(() => {
                    this.validateSwaggerSpec(swaggerSpec)
                        .then(() => {
                            this.parseSwaggerSpec(swaggerSpec)
                                .then((api) => {
                                    const apiInfo = api.info;
                                    const apiPaths = api.paths;
                                    const apiParameters = api.parameters;
                                    const apiDefinitions = api.definitions;
                                    const apiResponses = api.responses;
                                    const generatorProperties = api[CONSTANTS.NOVA.GENERATOR_PROPERTIES];

                                    winston.debug('############ INFORMATION ############');

                                    const apiModel = new ApiModel();
                                    apiModel.name = apiInfo.title;
                                    apiModel.description = apiInfo.description;
                                    apiModel.version = apiInfo.version;
                                    apiModel.basePath = api.basePath;
                                    apiModel.uuaa = generatorProperties
                                        ? generatorProperties[CONSTANTS.NOVA.BUSINESS_UNIT]
                                        : 'N/A';

                                    winston.debug('############ APIS ############');

                                    const paths = Object.keys(apiPaths);

                                    paths.forEach((pathParam) => {
                                        winston.debug(`Endpoint ${pathParam}`);
                                        const uri = pathParam;

                                        // 15/02/2018: Validate whether we're checking a socket or a normal path.
                                        const isSocket = apiPaths[pathParam][CONSTANTS.SOCKET.SOCKET_LABEL];

                                        if (isSocket) {
                                            const socket = this.extractSocketData(uri, apiPaths[pathParam]);
                                            apiModel.sockets.push(socket);
                                        } else {
                                            const methodsData = this.extractMethodsData(uri, apiPaths[pathParam]);
                                            apiModel.endpoints = apiModel.endpoints.concat(methodsData);
                                        }
                                    });

                                    winston.debug('############ PARAMETERS ############');

                                    if (apiParameters) {
                                        const parameters = Object.keys(apiParameters);
                                        parameters.forEach((parameter) => {
                                            const parameterData = apiParameters[parameter];
                                            winston.debug(`Parameter ${parameterData.name}`);

                                            const parameterModel = new ParameterModel();
                                            parameterModel.definitionName = parameter;
                                            parameterModel.name = parameterData.name;
                                            parameterModel.location = parameterData.in;
                                            parameterModel.description = parameterData.description;
                                            parameterModel.required = parameterData.required;

                                            // body parameters must define a schema
                                            // All other parameters use the type node instead
                                            // Also, if array type parameters must define an items node aswell
                                            if (parameterModel.location === 'body') {
                                                const schema = parameterData.schema;
                                                // 12/06/2018
                                                // Added schema support instead of simple reference support
                                                parameterModel.format = this.parseSchema(parameterData.name, schema);
                                            } else {
                                                parameterModel.format = this.getType(parameterData);
                                            }

                                            apiModel.parameters.push(parameterModel);
                                        });
                                    } else {
                                        winston.debug('No parameters found');
                                    }

                                    winston.debug('############ DEFINITIONS ############');

                                    if (apiDefinitions) {
                                        const definitions = Object.keys(apiDefinitions);
                                        definitions.forEach((definition) => {
                                            const definitionData = apiDefinitions[definition];
                                            winston.debug(`Definition ${definition}`);

                                            const definitionModel = new DefinitionModel();
                                            definitionModel.name = definition;

                                            // 12/06/2018
                                            // Support allOf on definitions
                                            if (definitionData.allOf) {
                                                const properties = definitionData.allOf;

                                                const attributes = [];
                                                Object.keys(properties).forEach((property) => {
                                                    const element = properties[property];

                                                    if (element.$ref) {
                                                        attributes.push({ extends: element.$ref });
                                                    } else {
                                                        const parsedSchema = this.parseSchema(
                                                            property,
                                                            properties[property]
                                                        );
                                                        // Cannot use concat because attribute is a constant
                                                        attributes.push.apply(attributes, parsedSchema.format);
                                                    }
                                                });
                                                // FIXME: This is required because of the nested format object on definitions
                                                definitionModel.format = { format: attributes };
                                            } else {
                                                definitionModel.format = this.parseSchema(definition, definitionData);
                                            }

                                            apiModel.definitions.push(definitionModel);
                                        });
                                    } else {
                                        winston.debug('No definitions found');
                                    }

                                    winston.debug('############ RESPONSES ############');

                                    if (apiResponses) {
                                        const responses = Object.keys(apiResponses);
                                        responses.forEach((response) => {
                                            const responseData = apiResponses[response];
                                            winston.debug(`Response ${response}`);

                                            const responseModel = new ResponseModel();
                                            responseModel.name = response;
                                            responseModel.description = responseData.description;
                                            responseModel.format = this.parseSchema(response, responseData.schema);

                                            apiModel.responses.push(responseModel);
                                        });
                                    } else {
                                        winston.debug('No responses found');
                                    }

                                    winston.debug('############ RESULT ############');
                                    resolve(apiModel);
                                })
                                .catch((err) => {
                                    reject(`Unable to parse API ${err}`);
                                });
                        })
                        .catch((err) => reject(err));
                })
                .catch((err) => reject(err));
        });
    }

    /**
     * <p>Method that extracts the input method's data.</p>
     *
     * @param uri {string} containing the path's URI
     * @param path {{}} containing the path data which methods we want to parse
     */
    extractMethodsData(uri, path) {
        const endpoints = [];

        const methods = Object.keys(path);

        methods
            .filter((method) => {
                return (
                    method === 'get' ||
                    method === 'post' ||
                    method === 'put' ||
                    method === 'options' ||
                    method === 'delete'
                );
            })
            .forEach((method) => {
                const endpointModel = new EndpointModel();
                // Use a RegExp here to support N-parameters on the endpoint definitions
                endpointModel.path = uri.replace(/{([^\\/]+)}/g, ':$1');
                endpointModel.method = method;
                endpointModel.description = path[method].description;
                endpointModel.summary = path[method].summary;
                endpointModel.functionToExec = this.allocateMockData(uri, method, path);

                if (path[method].parameters) {
                    path[method].parameters.forEach((parameter) => {
                        const parameterModel = new ParameterModel();

                        if (parameter.$ref) {
                            parameterModel.format = parameter.$ref;
                        } else {
                            parameterModel.name = parameter.name;
                            parameterModel.location = parameter.in;
                            parameterModel.description = parameter.description;
                            parameterModel.required = parameter.required;

                            // body parameters must define a schema
                            // All other parameters use the type node instead
                            // Also, if array type parameters must define an items node aswell
                            if (parameter.in === 'body') {
                                const schema = parameter.schema;
                                parameterModel.format = this.parseReference(schema);
                            } else {
                                parameterModel.format = this.getType(parameter);
                            }
                        }

                        endpointModel.parameters.push(parameterModel);
                    });
                }

                const responseCodes = Object.keys(path[method].responses);
                responseCodes.forEach((code) => {
                    const responseModel = new ResponseModel();
                    responseModel.code = code;

                    const format = path[method].responses[code];
                    responseModel.description = format.description;

                    if (format.$ref) {
                        responseModel.format = this.parseReference(format);
                    } else if (format.schema) {
                        const parsedSchema = this.parseSchema(code, format.schema);

                        if (parsedSchema.type === 'object') {
                            // TODO: Implement anonymous response objecs support?
                            throw new Error('Non-reference (anonymous) objects in responses is not supported');
                        }

                        responseModel.format = parsedSchema.type;
                    }

                    endpointModel.responses.push(responseModel);
                });

                endpoints.push(endpointModel);
            });

        return endpoints;
    }

    /**
     * <p>Method that parses the mocked responses yml</p>
     *
     * @param mockFilePath {string} path to file with mocks
     * @returns {Promise} resolves the generated <i>JSON</i> object or rejects with an error in case of error.
     */
    extractMockFileData(mockFilePath) {
        return new Promise((resolve, reject) => {
            if (!mockFilePath) {
                resolve();
            } else {
                fs.readFile(mockFilePath, (err, result) => {
                    if (err) {
                        winston.error(err);
                        reject(err);
                    } else {
                        try {
                            this.mockFileData = yamljs.parse(result.toString());
                            resolve();
                        } catch (err) {
                            winston.error(err);
                            reject(err);
                        }
                    }
                });
            }
        });
    }

    /**
     * <p>Method that extracts the input socket data.</p>
     *
     * @param uri {string} containing the path's URI
     * @param socket {{}} containing the socket data
     * @return {SocketModel} containing the socket information required for it's generation
     */
    extractSocketData(uri, socket) {
        const socketData = socket.get;

        const transport = socketData[CONSTANTS.SOCKET.TRANSPORT] || CONSTANTS.SOCKET.DEFAULT_TRANSPORT;
        const ping = socketData[CONSTANTS.SOCKET.PING] || CONSTANTS.SOCKET.DEFAULT_PING;

        const socketModel = new SocketModel();
        socketModel.description = socketData.description;
        socketModel.url = uri;
        socketModel.interval = socketData[CONSTANTS.SOCKET.INTERVAL];
        socketModel.format = socketData[CONSTANTS.SOCKET.FORMAT];
        socketModel.event = socketData[CONSTANTS.SOCKET.EVENT];
        socketModel.transport = transport;
        socketModel.ping = ping;
        // Sockets only support the GET method
        socketModel.functionToExec = this.allocateMockData(uri, 'get', socket);

        if (CONSTANTS.SOCKET.ALLOWED_TRANSPORTS.indexOf(transport) === -1) {
            throw new Error(`Unsupported transport '${transport}' for socket ${uri}.`);
        }

        if (transport === 'polling' && isNaN(ping)) {
            throw new Error('Ping should be a defined number for polling transports.');
        }

        return socketModel;
    }

    /**
     * <p>Method that parses the input <code>Type</code> object.</p>
     *
     * @param object {{}} containing a <i>Swagger</i> spec type node.
     */
    getType(object) {
        let type = object.type;
        if (object.type === 'array') {
            const parsedType = this.parseArrayItems(object);
            type += `[${parsedType}]`;
        }

        return type;
    }

    /**
     * <p>Method that parses the input <code>Array</code> object.</p>
     *
     * @param array {{}} containing a <i>Swagger</i> spec array node.
     */
    parseArrayItems(array) {
        let arrayType;
        if (array.items.$ref) {
            arrayType = array.items.$ref;
        } else if (array.items.type === 'object') {
            // TODO: Implement anonymous response objecs support?
            // const arrayItemSchema = this.parseSchema('arrayItem', array.items);
            throw new Error('Non-reference (anonymous) objects in array definitions is not supported');
        } else {
            arrayType = array.items.type;
        }

        return `${arrayType}`;
    }

    /**
     * <p>Method that parses the input <code>Reference</code> object.</p>
     *
     * @param reference {{}} containing a <i>Swagger</i> spec reference node.
     */
    parseReference(reference) {
        let schemaType;
        if (reference.$ref) {
            schemaType = reference.$ref;
        } else {
            schemaType = this.getType(reference);
        }

        return schemaType;
    }

    /**
     * <p>Method that parses the input <code>Schema</code> object.</p>
     *
     * @param name {string} with the schema parent name.
     * @param schema {SchemaModel} containing a <i>Swagger</i> spec schema node.
     */
    parseSchema(name, schema) {
        const schemaType = new SchemaModel();
        if (schema.type === 'object') {
            if (schema.properties) {
                const properties = schema.properties;

                const attributes = [];
                Object.keys(properties).forEach((property) => {
                    attributes.push(this.parseSchema(property, properties[property]));
                });

                const type = this.getType(schema);

                schemaType.name = name;
                schemaType.type = type;
                schemaType.format = attributes;
            } else if (schema.additionalProperties) {
                const properties = schema.additionalProperties;

                // Type is the node used by the parser to actually fetch the format
                // A single model with a name 'additionalProperty' with the correct type is created
                const schemaFormat = new SchemaModel();
                schemaFormat.name = 'additionalProperty';
                schemaFormat.type = this.getType(properties);

                schemaType.name = 'additionalProperties';
                schemaType.type = 'object';
                schemaType.format = [schemaFormat];
            } else if (schema.allOf) {
                const properties = schema.allOf;

                const attributes = [];
                Object.keys(properties).forEach((property) => {
                    const element = properties[property];

                    if (element.$ref) {
                        attributes.push({ extends: element.$ref });
                    } else {
                        const parsedSchema = this.parseSchema(property, properties[property]);
                        // Cannot use concat because attribute is a constant
                        attributes.push.apply(attributes, parsedSchema.format);
                    }
                });

                const type = this.getType(schema);

                schemaType.name = name;
                schemaType.type = type;
                schemaType.format = attributes;
            }
        } else if (schema.$ref) {
            schemaType.name = name;
            schemaType.type = schema.$ref;
        } else {
            const type = this.getType(schema);

            schemaType.name = name;
            schemaType.description = schema.description;
            schemaType.type = type;
            schemaType.required = schema.required;
        }

        return schemaType;
    }

    /**
     * <p>Method that parses the given <i>Swagger</i> spec and transforms it into a <i>JSON</i> object.
     * <b>NOTE</b>: The file is bundled so object references are kept by name.</p>
     *
     * @param swaggerSpec {{}} <code>YAML</code> <i>Swagger</i> spec file.
     * @returns {Promise} resolves the generated <i>JSON</i> object or rejects with an error in case of error.
     */
    parseSwaggerSpec(swaggerSpec) {
        return new Promise((resolve, reject) => {
            const YAML = SwaggerParser.YAML;
            const jsonSpec = YAML.parse(swaggerSpec);

            SwaggerParser.bundle(jsonSpec)
                .then((api) => {
                    resolve(api);
                })
                .catch((err) => {
                    reject(err);
                });
        });
    }

    /**
     * <p>Method that validates the input <i>Swagger</i> spec.</p>
     *
     * @param swaggerSpec {{}} <code>YAML</code> <i>Swagger</i> spec file.
     * @returns {Promise} resolves if the <i>Swagger</i> spec is correctly validates or rejects with the error
     * message in case or error.
     */
    validateSwaggerSpec(swaggerSpec) {
        return new Promise((resolve, reject) => {
            const YAML = SwaggerParser.YAML;
            const jsonSpec = YAML.parse(swaggerSpec);

            SwaggerParser.validate(jsonSpec)
                .then(() => {
                    resolve();
                })
                .catch((err) => {
                    reject(err);
                });
        });
    }
}

module.exports = new SwaggerApiParser();
