'use strict';

const winston = require('winston');

/**
 * <p>Class that validates the configured parameter.</p>
 */
class ParameterValidator {
    /**
     * <p><code>Validator</code> constructor which initializes it's configuration.</p>
     *
     * @param name {string} with the parameter name.
     * @param location {string} with the parameter location.
     * @param format {{}|string} with the parameter format.
     */
    constructor(name, location, format) {
        this._name = name;
        this._location = location;
        this._format = format;
    }

    /**
     * <p>Parameter name getter</p>
     * @returns {string|*} with the parameter name.
     */
    get name() {
        return this._name;
    }

    /**
     * <p>Method that performs the given object validation.</p>
     *
     * @param content {{}} with the object parameter content.
     * @param type {string} with the object type.
     * @returns {boolean} <code>true</code> if the parameter is valid, <code>false</code> otherwise.
     * @private
     */
    _validateObject(content, type) {
        winston.debug(`Validating object ${JSON.stringify(content)}`);
        let validObject = true;

        // Validate that the object has all the defined keys
        const keys = Object.keys(type);
        keys.forEach((key) => {
            const keyValue = content[key];
            if (typeof keyValue === 'object') {
                const validKey = this._validateObject(keyValue, type[key]);
                validObject = validObject || validKey;
                winston.debug(`Key ${key} is ${validKey ? 'a valid' : 'an invalid'} object`);
            } else {
                const validKey = this._validatePrimitive(keyValue, type[key]);
                validObject = validObject || validKey;
                winston.debug(`Key ${key} is ${validKey ? 'a valid' : 'an invalid'} primitive`);
            }
        });

        return validObject;
    }

    /**
     * <p>Method that performs the given primitive type validation.</p>
     *
     * @param content {string} with the primitive type parameter content.
     * @param type {string} with the primitive type.
     * @returns {boolean} <code>true</code> if the parameter is valid, <code>false</code> otherwise.
     * @private
     */
    _validatePrimitive(content, type) {
        switch (type) {
            case 'string':
                winston.debug(`Validating string: ${this._validateString(content)}`);
                return this._validateString(content);
            case 'number':
            case 'integer':
                winston.debug(`Validating number: ${this._validateNumber(content)}`);
                return this._validateNumber(content);
            case 'boolean':
                winston.debug(`Validating boolean: ${this._validateBoolean(content)}`);
                return this._validateBoolean(content);
            default:
                return false;
        }
    }

    /**
     * <p>Method that checks if the input parameter content is a <code>string</code>.</p>
     *
     * @param expectedString {string} with the parameter content.
     * @returns {boolean} <code>true</code> if the input content is a <code>string</code>,
     * <code>false</code> otherwise.
     * @private
     */
    _validateString(expectedString) {
        return typeof expectedString === 'string';
    }

    /**
     * <p>Method that checks if the input parameter content is a <code>number</code>.</p>
     *
     * @param expectedNumber {string} with the parameter content.
     * @returns {boolean} <code>true</code> if the input content is a <code>number</code>,
     * <code>false</code> otherwise.
     * @private
     */
    _validateNumber(expectedNumber) {
        return !isNaN(Number(expectedNumber));
    }

    /**
     * <p>Method that checks if the input parameter content is a <code>boolean</code>.</p>
     *
     * @param expectedBoolean {string} with the parameter content.
     * @returns {boolean} <code>true</code> if the input content is a <code>boolean</code>,
     * <code>false</code> otherwise.
     * @private
     */
    _validateBoolean(expectedBoolean) {
        return typeof expectedBoolean === 'boolean';
    }

    /**
     * <p>Method that validates the configured parameter.</p>
     *
     * @param request {request} object received by the <code>Express</code> endpoint.
     * @returns {boolean} <code>true</code> if the parameter content is valid according to the
     * configuration.
     */
    validate(request) {
        let validParameter = true;
        winston.debug(
            `Validating: ${this._name} which is a ${this._location} parameter with format ${JSON.stringify(
                this._format
            )}`
        );
        const body = request.body;
        const headerParam = request.header(this._name);
        const pathParam = request.params[this._name];
        const queryParam = request.query[this._name];
        const formData = request.body[this._name];
        switch (this._location) {
            case 'cookie':
            case 'body':
                winston.debug(`Content ${JSON.stringify(body)}`);
                if (typeof this._format === 'object') {
                    validParameter = this._validateObject(body, this._format);
                } else {
                    validParameter = this._validatePrimitive(body, this._format);
                }
                break;
            case 'header':
                winston.debug(`Content ${JSON.stringify(headerParam)}`);
                validParameter = this._validatePrimitive(headerParam, this._format);
                break;
            case 'path':
                winston.debug(`Content ${JSON.stringify(pathParam)}`);
                validParameter = this._validatePrimitive(pathParam, this._format);
                break;
            case 'query':
                winston.debug(`Content ${JSON.stringify(queryParam)}`);
                validParameter = this._validatePrimitive(queryParam, this._format);
                break;
            case 'formData':
                winston.debug(`Content ${JSON.stringify(formData)}`);
                validParameter = this._validatePrimitive(formData, this._format);
                break;
            default:
                winston.error(`Unable to resolve parameter ${this._name} location`);
                validParameter = false;
        }

        return validParameter;
    }
}

module.exports = ParameterValidator;
