'use strict';

const winston = require('winston');

/**
 * <p>Class that generates the configured response.</p>
 */
class ResponseGenerator {
    /**
     * <p><code>Generator</code> constructor which initializes it's configuration.</p>
     *
     * @param code {string} the <i>HTTP</i> response code.
     * @param format {{}|string} the response format.
     */
    constructor(code, format) {
        this.code = code;
        this._format = format;
        this._formatResponse.bind(this);
        this.generate.bind(this);
    }

    /**
     * <p>Method that handles array content generation from the given content definition.</p>
     *
     * @param {{}} content contains the array definition to parse and generate.
     * @returns {*} with the array response generated according to the given definition.
     * @private
     */
    _handleArray(content) {
        // If it's an array, format it's elements
        winston.debug(`${JSON.stringify(content)}`);
        // 15/02/2018: This array should only have one element, which is the object definition
        // 15/06/2018 Don't format the response when the response is a primitive type :)
        let elementFormat;
        const dataType = content[0];
        if (this._isPrimitive(dataType)) {
            elementFormat = this._handlePrimitive(dataType);
        } else {
            elementFormat = this._formatResponse(content[0]);
        }

        // TODO 15/02/2018: Allow the configuration of this generation by some custom property?
        // By default, generate 20 elements of the retrieved format
        const arrayContent = [elementFormat];
        for (let noElement = 1; noElement < 20; noElement++) {
            arrayContent.push(elementFormat);
        }

        return arrayContent;
    }

    /**
     * <p>Method that handles object content generation from the given content definition.</p>
     *
     * @param {{}} content contains the object definition to parse and generate.
     * @returns {*} with the object response generated according to the given definition.
     * @private
     */
    _handleObject(content) {
        const objectContent = {};
        const contentKeys = Object.keys(content);
        contentKeys.forEach((key) => {
            if (Array.isArray(content[key])) {
                objectContent[key] = this._handleArray(content[key]);
            } else if (typeof content[key] === 'object') {
                // If it's a json object, format it's properties
                objectContent[key] = this._formatResponse(content[key]);
            } else {
                // If it's a primitive type, change for proper values
                objectContent[key] = this._handlePrimitive(content[key]);
            }
        });
        return objectContent;
    }

    /**
     * <p>Method that handles primitive type content generation from the given content definition.</p>
     *
     * @param type {string} primitive data type to generate a value for (supports number, boolean or string)
     * @returns {number|boolean|string} generated primitive data type
     * @private
     */
    _handlePrimitive(type) {
        let content;
        switch (type) {
            case 'number':
            case 'integer':
                content = 0;
                break;
            case 'boolean':
                content = true;
                break;
            case 'string':
                content = 'string';
                break;
            default:
                throw Error(`Invalid primitive type: ${type}`);
        }
        return content;
    }

    /**
     * <p>Method that validates whether the input type is a primitive type or not.</p>
     *
     * @param {string} type an object type.
     * @returns {boolean} true if the type is primitive, false otherwise.
     * @private
     */
    _isPrimitive(type) {
        return type === 'string' || type === 'number' || type === 'integer' || type === 'boolean';
    }

    /**
     * <p>Method that formats the response object generated according to the configured format.
     * For example, transforms { 'test': 'integer' } into { 'test': 0 }.</p>
     * <p>12/06/2018 - fixed a bug that caused the format to the corrupted on subsequent calls.</p>
     * <p>24/03/2019 - Added primitive types support.</p>
     *
     * @param content {{}|Array|string} the object format.
     * @returns {*} the formatted object
     * @private
     */
    _formatResponse(content) {
        let responseContent;

        if (Array.isArray(content)) {
            responseContent = this._handleArray(content);
        } else if (this._isPrimitive(content)) {
            responseContent = this._handlePrimitive(content);
        } else {
            responseContent = this._handleObject(content);
        }

        return responseContent;
    }

    /**
     * <p>Method that generates a response object and uses the <i>ExpressJS</i> response to deliver it.</p>
     *
     * @param response {response} the <i>ExpressJS</i> response.
     */
    generate(response) {
        winston.debug(`Generating response ${this.code}:${JSON.stringify(this._format)}`);

        if (this.code !== 'default') {
            response.status(this.code);
        }

        if (this._format) {
            const responseBody = this._formatResponse(this._format);
            response.json(responseBody);
        } else {
            response.send();
        }
    }
}

module.exports = ResponseGenerator;
